/* Crazy Crane */

#include "main.h"
#include "gfx.h"
#include "random.h"
#include "counter.h"

#include <proto/exec.h>
#include <proto/graphics.h>
#include <proto/intuition.h>


// The game playfield. Lower byte of every field contains letter. Special fields are destignated as follows:
// ' ' - background
// '[' - bar left edge
// ']' - bar right edge
// '-' - bar internal element
// Upper byte is for flags:


extern struct GameData Gd;


struct QuestionEntry
{
	struct MinNode Node;
	STRPTR Phrase;
	STRPTR Category;
};


struct LetterGroup
{
	UBYTE *Letters;
	WORD Length;
};

/* For Polish alphabet. No more than 4 letters in group are allowed. Graphics layout is broken otherwise. */

struct LetterGroup LetterGroups[] = {
	{ (UBYTE*)"ABC", 4 },
	{ (UBYTE*)"DE", 4 },
	{ (UBYTE*)"FGHI", 4 },
	{ (UBYTE*)"JKL", 3 },
	{ (UBYTE*)"MN", 4 },
	{ (UBYTE*)"OPQ", 4 },
	{ (UBYTE*)"RST", 4 },
	{ (UBYTE*)"UVWX", 4 },
	{ (UBYTE*)"YZ", 4 },
	{ NULL, 0 }
};

UBYTE Permutations3[6][3] = {
	{ 0, 1, 2 }, { 0, 2, 1 }, { 1, 0, 2 }, { 1, 2, 0 }, { 2, 0, 1 }, { 2, 1, 0 }
}; 

UBYTE Permutations4[24][4] = {
	{ 0, 1, 2, 3 }, { 0, 1, 3, 2 }, { 0, 2, 1, 3 }, { 0, 2, 3, 1 }, { 0, 3, 1, 2 }, { 0, 3, 2, 1 },
	{ 1, 0, 2, 3 }, { 1, 0, 3, 2 }, { 1, 2, 0, 3 }, { 1, 2, 3, 0 }, { 1, 3, 0, 2 }, { 1, 3, 2, 0 },
	{ 2, 0, 1, 3 }, { 2, 0, 3, 1 }, { 2, 1, 0, 3 }, { 2, 1, 3, 0 }, { 2, 3, 0, 1 }, { 2, 3, 1, 0 },
	{ 3, 0, 1, 2 }, { 3, 0, 2, 1 }, { 3, 1, 0, 2 }, { 3, 1, 2, 0 }, { 3, 2, 0, 1 }, { 3, 2, 1, 0 }
}; 

// The first index is index of clicked block, which stays, and then falls down.

UBYTE VanishOrder[4][3] = {
	{ 3, 2, 1 }, { 0, 3, 2 }, { 0, 1, 3 }, { 0, 1, 2 }
};


//==============================================================================================================================
// Converts mouse x coordinate from window pixels to blocks. It assumes that Intuition only reports events over window inside.
// WORD/UWORD type juggling avoids linking with __divsi3 on m68k.

WORD GetMouseX(WORD x)
{
	UWORD x1 = (UWORD)x;

	x1 -= Gd.GameWin->BorderLeft;
	x1 /= (UWORD)Gd.BlockWidth;
	return (WORD)x1;
}


//==============================================================================================================================
// Converts mouse y coordinate from window pixels to blocks. It assumes that Intuition only reports events over window inside.
// WORD/UWORD type juggling avoids linking with __divsi3 on m68k.

WORD GetMouseY(WORD y)
{
	UWORD y1 = (UWORD)y;

	y1 -= Gd.GameWin->BorderTop;
	y1 /= (UWORD)Gd.BlockHeight;
	return (WORD)y1;
}


//==============================================================================================================================
// Given block coordinate of a clicked block, the function unclicks all blocks in the same letter column.

void UnclickInColumn(UWORD x, UWORD y)
{
	UWORD y_start, y_unclick, field;

	if (y < 5) y_start = 1;
	else if (y < 10) y_start = 6;
	else y_start = 11;

	for (y_unclick = y_start; y_unclick < y_start + 4; y_unclick++)
	{
		field = Gd.Field[y_unclick * FIELD_WIDTH + x];

		if ((field & FF_CLICKED) && (y_unclick != y)) 
		{
			field ^= FF_CLICKED;
			field |= FF_CLICKABLE;
			BlitLetter(field & FF_LETTERMASK, x, y_unclick, FALSE);
			Gd.Field[y_unclick * FIELD_WIDTH + x] = field;
		}
	}
}


//==============================================================================================================================
// After a letter is clicked at (x, y), the function updates ClickedString with clicked letter. Returns letter index in Clicked
// String.

UWORD UpdateClickedString(UWORD x, UWORD y)
{
	UWORD letter_index = 0;

	if (y > 5) letter_index += Gd.LettersInRow[0] + 1;      // + 1 is for space between this and next row
	if (y > 10) letter_index += Gd.LettersInRow[1] + 1;
	letter_index += x - 1;
	Gd.ClickedString[letter_index] = Gd.Field[y * FIELD_WIDTH + x] & FF_LETTERMASK;
	return letter_index;
}


//==============================================================================================================================
// This function receives the range (first letter, number of letters) occupied by just guessed word in the phrase. For every
// column of this range, not clicked letters disappear one by one, then clicked letter is scrolled down to drop onto the 
// bar. Then "barquake" effect is performed.

void DropGuessedWordDown(UWORD first_letter, UWORD letters)
{
	WORD i, display_row, display_column, logical_row = 0;

	/*------------------------------------------------*/
	/* Calculate phrase row and phrase column in row. */
	/*------------------------------------------------*/

	if (first_letter >= Gd.LettersInRow[0] + 1)
	{
		first_letter -= Gd.LettersInRow[0] + 1;
		logical_row = 1;

		if (first_letter >= Gd.LettersInRow[1] + 1)
		{
			first_letter -= Gd.LettersInRow[1] + 1;
			logical_row = 2;
		}
	}

	/*---------------------------------------------------*/
	/* Loop over columns of the word and make unclicked  */
	/* letters in each column to disappear one by one.   */
	/*---------------------------------------------------*/

	display_row = logical_row * 5 + 1;   // upper row of letters for logical row

	for (display_column = first_letter + 1; display_column <= first_letter + letters; display_column++)
	{
		WORD scroll_rows = 0;
		WORD clicked_letter_row = 0;
		UWORD *field = &Gd.Field[display_row * FIELD_WIDTH + display_column];
		UWORD *f2;
		UWORD clicked_letter = 0;

		f2 = field;

		/* Finding row of clicked (guessed) letter, counting from the top row. */

		for (i = 0; i < 4; i++)
		{
			if (*f2 & FF_CLICKED)
			{
				clicked_letter = *f2 & ~FF_CLICKABLE;
				clicked_letter_row = i;
				scroll_rows = 3 - i;
				break;
			}

			f2 += FIELD_WIDTH;
		}

		/* Vanishing unclicked letters in physically correct order. */

		for (i = 0; i < 3; i++)
		{
			UWORD v = VanishOrder[clicked_letter_row][i];

			f2 = field + v * FIELD_WIDTH;

			if (*f2 & FF_CLICKABLE)
			{
				struct AnimEvent *aev;

				if (aev = AllocTaskPooled(sizeof(struct AnimEvent)))
				{
					aev->EventKind = ANIM_EVENT_CLEAR;
					aev->NextGameState = 0;
					aev->WaitAfter = 6;
					aev->Points = letters + 5;
					aev->X0 = display_column;
					aev->Y0 = display_row + v;
					AddTail((struct List*)&Gd.AnimQueue, (struct Node*)aev);
				}
			}
		}

		/* Perform falling animations. */

		if (scroll_rows > 0)   // no scroll and bar quake if letter already rests on the bar
		{
			GenerateAnimColumnDown(display_column, display_row + 3, scroll_rows, clicked_letter);
			GenerateAnimBarQuake(logical_row);
		}
	}
}

//==============================================================================================================================

BOOL PhraseIsGuessed(void)
{
	if (StrEqu(Gd.OriginalString, Gd.ClickedString)) return TRUE;
	else return FALSE;
}


//==============================================================================================================================
// After click at given block (x, y) the function checks if the word, where click was located, is guessed or not. If the word
// is guessed, routine for dropping letters down is called.

BOOL WordIsGuessed(UWORD x, UWORD y)
{
	UWORD letter_index, word_start, word_length;

	letter_index = UpdateClickedString(x, y);

	/*---------------------------------------------------------------------------------------*/
	/* Go back either to a space before the clicked word, or to the start of guessed phrase. */
	/*---------------------------------------------------------------------------------------*/

	for (word_start = letter_index; word_start > 0; word_start--)
	{
		if (Gd.OriginalString[word_start - 1] == ' ') break;
	}

	/*-----------------------------*/
	/* Now count the word letters. */
	/*-----------------------------*/

	for (word_length = 0; Gd.OriginalString[word_start + word_length]; word_length++)
	{
		if (Gd.OriginalString[word_start + word_length] == ' ') break;
	}

	/*---------------------------------------------------------------------*/
	/* Compare word in OriginalString with the same word in ClickedString. */
	/* If they match, the word has been guessed.                           */
	/*------------------------------------------ --------------------------*/

	for (letter_index = word_start; letter_index < word_start + word_length; letter_index++)
	{
		if (Gd.OriginalString[letter_index] != Gd.ClickedString[letter_index]) return FALSE;
	}

	/*--------------------------------------------------*/
	/* Generate and push AnimEvents into the AnimQueue. */
	/*--------------------------------------------------*/

	DropGuessedWordDown(word_start, word_length);
	return TRUE;
}


//==============================================================================================================================
// Function returns game state after click. It is usually GAME_SOLVING_PUZZLE, unless the whole phrase has been just guessed.

UWORD HandleClick(UWORD x, UWORD y)
{
	UWORD field, result = GAME_SOLVING_PUZZLE;

	field = Gd.Field[y * FIELD_WIDTH + x];

	if (field & FF_CLICKABLE)
	{
		field ^= FF_CLICKED | FF_CLICKABLE;
		BlitLetter(field & FF_LETTERMASK, x, y, field & FF_CLICKED);
		Gd.Field[y * FIELD_WIDTH + x] = field;
		UnclickInColumn(x, y);

		if (WordIsGuessed(x, y))
		{
			if (PhraseIsGuessed())
			{
				//FadeToBackground(Win, GameGfx);
				result = 7;
			}
		}
	}

	return result;
}


//==============================================================================================================================
// Blits the whole field. Used for displaying new puzzle. This function assumes the background is not damaged, so it does not
// blit empty blocks.

void BlitField(void)
{
	UWORD x, y, field;

	for (y = 0; y < FIELD_HEIGHT; y++)
	{
		for (x = 0; x < FIELD_WIDTH; x++)
		{
			field = Gd.Field[y * FIELD_WIDTH + x];

			if (field != ' ') BlitLetter(field & FF_LETTERMASK, x, y, field & FF_CLICKED);
		}
	}
}


//==============================================================================================================================
// Finds a group (in 'LetterGroups'), a letter belongs to. Returns pointer to LetterGroup structure. Returns NULL if letter is
// not found in any group.

struct LetterGroup *GetLetterGroup(char letter)
{
	struct LetterGroup *lg = LetterGroups;
	char *p, x;

	while (p = (char*)lg->Letters)
	{
		while (x = *p++)
		{
			if (x == letter) return lg;
		}

		lg++;
	}

	return NULL;
}


//==============================================================================================================================
// Fills the whole playfield with background.

void ErasePlayfield(void)
{
	UWORD i;

	for (i = 0; i < FIELD_WIDTH * FIELD_HEIGHT; i++) Gd.Field[i] = ' ';
}


//==============================================================================================================================
// Function puts a column of letters for given question letter, number of row and x position in row.

void PutColumn(char letter, UWORD row, UWORD x)
{
	struct LetterGroup *lg;
	UWORD y, i, field;
	UBYTE perm_index;

	/* Putting bar element first. The first column gets starting bar element. */

	y = 5 * (row + 1);
	field = '-';
	if (x == 1) field = '[';
	Gd.Field[y * FIELD_WIDTH + x] = field;

	/* If letter is found in groups, put a stack of letters of the group in random order. */

	if (lg = GetLetterGroup(letter))
	{
		y -= lg->Length;
		perm_index = RandomScaled(23);
		if (lg->Length == 3) perm_index /= 6;

		for (i = 0; i < lg->Length; i++)
		{
			if (lg->Length == 3) field = lg->Letters[Permutations3[perm_index][i]];
			else field = lg->Letters[Permutations4[perm_index][i]];
			Gd.Field[y * FIELD_WIDTH + x] = field | FF_CLICKABLE;
			y++;
		}
	}
}

//==============================================================================================================================

void PutWord(char *s, UWORD row, UWORD wlen, UWORD lcount)
{
	while (wlen--) PutColumn(*s++, row, 1 + lcount++);
}

  
//==============================================================================================================================
// Returns length of word in the sentence.

WORD WordLength(char *s)
{
	char *p = s;

	while (*s && (*s != ' ')) s++;
	return (WORD)(s - p);
}

//==============================================================================================================================
// Decodes a string containing the sentence to guess. Fills game playfield with objects (letters and bars).

void PrepareQuestion(void)
{
	char *s = Gd.OriginalString;
	WORD wlen, lcount, row;

	Gd.LettersInRow[0] = 0;
	Gd.LettersInRow[1] = 0;
	Gd.LettersInRow[2] = 0;
	row = 0;
	lcount = 0;
	ErasePlayfield();

	while (wlen = WordLength(s))
	{
		if (lcount == 0)     // the first word in a row
		{
			if (wlen <= 28)
			{
				PutWord(s, row, wlen, lcount);
				lcount += wlen;
				s += wlen;
				if (*s) s++;      // skip leading space
			}
			else { PutStr("Sowo za dugie.\n"); break; }
		}
		else
		{
			if (lcount + wlen + 1 <= 28)
			{
				PutColumn(' ', row, lcount + 1);
				lcount++;
				PutWord(s, row, wlen, lcount);
				lcount += wlen;
				s += wlen;
				if (*s) s++;
			}
			else
			{
				Gd.LettersInRow[row] = lcount;
				row++;
				Gd.Field[5 * row * FIELD_WIDTH + lcount] = ']';
				lcount = 0;
				if (row > 2) { PutStr("Haso za dugie\n"); break; }
			}
		}
	}

	Gd.Field[5 * (row + 1) * FIELD_WIDTH + lcount] = ']';
	Gd.LettersInRow[row] = lcount;
}


//==============================================================================================================================
// Replaces all letters in a string with question marks (marks them as "no letter clicked in a column"). Spaces are left intact.

void UnclickClickedString()
{
	STRPTR s = Gd.ClickedString;
	char c;

	while (c = *s)
	{
		if (c != ' ') *s = '?';
		s++;
	}
}

//==============================================================================================================================
// Checks point decrease counter. If counts down to 0, decreases and updates points, restarts counter.

void DecreasePoints(void)
{
	if (Gd.PointsCountDown == 0)
	{
		Gd.PointsCountDown = POINT_INTERVAL;

		if (Gd.Points > 0)
		{
			Gd.Points--;
			Printf("Points: %5ld\r", Gd.Points);
		}
	}
	else Gd.PointsCountDown--;
}

//==============================================================================================================================

void StateTransition(UWORD old_state, UWORD new_state)
{
	if (new_state == GAME_SOLVING_PUZZLE)
	{
		PrepareQuestion();
		BlitField();
	}
}

//==============================================================================================================================

UWORD HandleAnimations(UWORD state)
{
	if (Gd.AnimCounter == 0)
	{
		struct AnimEvent *aev;

		if (Gd.DelayedState != 0)
		{
			StateTransition(state, Gd.DelayedState);
			state = Gd.DelayedState;
			Gd.DelayedState = 0;
		}

		if (aev = (struct AnimEvent*)RemHead((struct List*)&Gd.AnimQueue))
		{
			Gd.DelayedState = aev->NextGameState;
			Gd.AnimCounter = aev->WaitAfter;

			switch (aev->EventKind)
			{
				case ANIM_EVENT_SCROLL:
					ScrollWindowRaster(Gd.GameWin, aev->X2, aev->Y2, aev->X0, aev->Y0, aev->X1, aev->Y1);
				break;

				case ANIM_EVENT_CLEAR:
					ClearLetter(aev->X0, aev->Y0);
					Gd.Field[aev->Y0 * FIELD_WIDTH + aev->X0] = ' ';
					Gd.Points += aev->Points;
					Printf("Points: %5ld\r", Gd.Points);
				break;

				case ANIM_EVENT_COLDOWN:
					PerformAnimColumnDown(aev->X0, aev->Y0, aev->Y1, aev->Letter);

					/* Update playfield. */

					if (aev->X2)
					{
						UWORD k = aev->Y2 * FIELD_WIDTH + aev->X1;

						Gd.Field[k - 3 * FIELD_WIDTH] = ' ';
						Gd.Field[k - 2 * FIELD_WIDTH] = ' ';
						Gd.Field[k - FIELD_WIDTH] = ' ';
						Gd.Field[k] = aev->Letter;
					}
				break;

				case ANIM_EVENT_BARQUAKE:
					PerformAnimBarQuake(aev->X0, aev->Y0, aev->Y1, aev->Y2);
				break;
			}

			FreeTaskPooled(aev, sizeof(struct AnimEvent));
		}
	}
	else Gd.AnimCounter--;

	return state;
}


//==============================================================================================================================

void EventLoop(void)
{
	BOOL running = TRUE;
	ULONG signals, winmask, ctrmask, state = GAME_TITLE_SCREEN;

	winmask = 1 << Gd.GameWin->UserPort->mp_SigBit;
	ctrmask = 1 << Gd.PCounter.Port->mp_SigBit;

	while (running)
	{
		signals = Wait(winmask | ctrmask | SIGBREAKF_CTRL_C);

		if (signals & winmask)
		{
			struct IntuiMessage *imsg;

			while (imsg = (struct IntuiMessage*)GetMsg(Gd.GameWin->UserPort))
			{
				if (imsg->Class == IDCMP_CLOSEWINDOW) running = FALSE;
				else if (imsg->Class == IDCMP_MOUSEBUTTONS)
				{
					if (imsg->Code == IECODE_LBUTTON)
					{
						switch (state)
						{
							case GAME_TITLE_SCREEN:
								state = GAME_CURTAIN_UP;
								AnimCurtainUp();
							break;

							case GAME_SOLVING_PUZZLE:
								state = HandleClick(GetMouseX(imsg->MouseX), GetMouseY(imsg->MouseY));
							break;
						}
					}
				}
			}
		}

		if (signals & ctrmask)      /* Game heartbeat (1/50 of second) */
		{
			state = HandleAnimations(state);
			DecreasePoints();
			CounterRetrig(&Gd.PCounter);
		}

		if (signals & SIGBREAKF_CTRL_C) running = FALSE;
	}
}


//==============================================================================================================================

#define LQ_PHRASE        1
#define LQ_CAT_SEARCH    2
#define LQ_CATEGORY      3
#define LQ_PHR_SEARCH    4

BOOL LoadQuestions(STRPTR filepath)
{
	BOOL success = FALSE;
	ULONG data_length, state = LQ_PHRASE;
	BPTR file_handle;
	STRPTR data_buffer, phrase = NULL, category = NULL;

	if (file_handle = Open(filepath, MODE_OLDFILE))
	{
		if (Seek(file_handle, 0, OFFSET_END) == 0)
		{
			if ((data_length = Seek(file_handle, 0, OFFSET_BEGINNING)) > 0)
			{
				if (data_buffer = AllocVecTaskPooled(data_length))    // will be disposed with the pool
				{
					if (Read(file_handle, data_buffer, data_length) == (LONG)data_length)
					{
						STRPTR s = data_buffer;
						STRPTR data_end = data_buffer + data_length;
						struct QuestionEntry *qen;
						BOOL error = FALSE;

						while (!error && (s < data_end))
						{
							switch (state)
							{
								case LQ_PHRASE:
									phrase = s;
									state = LQ_CAT_SEARCH;
								break;

								case LQ_CAT_SEARCH:
									if (*s == '|')
									{
										*s++ = 0x00;
										state = LQ_CATEGORY;
									}
									else s++;
								break;

								case LQ_CATEGORY:
									category = s;
									state = LQ_PHR_SEARCH;
								break;

								case LQ_PHR_SEARCH:
									if (*s == 0x0A)   // end of line
									{
										*s++ = 0x00;

										if (phrase && category && (qen = AllocTaskPooled(sizeof(struct QuestionEntry))))
										{
											qen->Phrase = phrase;
											qen->Category = category;
											AddTail((struct List*)&Gd.Questions, (struct Node*)qen);
											Gd.QuestionCount++;
											state = LQ_PHRASE;
										}
										else error = TRUE;
									}
									else s++;
								break;
							}
						}

						if (!error) success = TRUE;
					}

					if (!success) FreeVecTaskPooled(data_buffer);
				}
			}
		}

		Close(file_handle);
	}

	return success;
}

//==============================================================================================================================

BOOL GetRandomQuestion(void)
{
	ULONG q, r = 0;
	struct QuestionEntry *qen;
	STRPTR wintitle = Gd.WinTitleBuf;

	q = RandomScaled(Gd.QuestionCount - 1);

	ITERATE_LIST(&Gd.Questions, struct QuestionEntry*, qen)
	{
		if (r++ == q)
		{
			Gd.OriginalString = qen->Phrase;

			if (Gd.ClickedString = StrNew(qen->Phrase))
			{
				wintitle = StrCopy("CrazyCrane: ", wintitle);
				if (StrLen(qen->Category) < 115) StrCopy(qen->Category, wintitle);
				SetWindowTitles(Gd.GameWin, Gd.WinTitleBuf, (CONST_STRPTR)~0);
				Remove((struct Node*)qen);
				Gd.QuestionCount--;
				return TRUE;
			}
			else break;
		}
	}

	return FALSE;
}

//==============================================================================================================================

void TheGame(void)
{
	INIT_LIST(&Gd.AnimQueue);
	Gd.AnimCounter = 0;
	INIT_LIST(&Gd.Questions);
	Gd.QuestionCount = 0;

	if (LoadQuestions("PROGDIR:pytania_aktualne") || LoadQuestions("PROGDIR:pytania_zestaw_1"))
	{
		if (CounterNew(&Gd.PCounter))
		{
			RandomSeed();
			BlitTitleGfx();
			if (GetRandomQuestion())
			{
				UnclickClickedString();
				EventLoop();
				//SaveQuestions("PROGDIR:pytania_aktualne");
				StrFree(Gd.ClickedString);
			}

			CounterDispose(&Gd.PCounter);
		}
	}
}

//==============================================================================================================================
